#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <errno.h>
#include <pwd.h>
#include <grp.h>
#include <getopt.h>

struct passwd *pw;
struct group *gr, *gr2;
int uid, gid, euid, egid;

struct passwd *
xgetpwuid(int u)
{
	struct passwd *p = getpwuid(u);
	
	if (p == NULL) {
		printf("Record not found or an error.\n");
		exit(1);
	}
	
	return p;
}

struct group *
xgetgrgid(int gd)
{
	struct group *g = getgrgid(gd);
	
	if (g == NULL) {
		printf("Record not found or an error.\n");
		exit(1);
	}
	
	return g;
}

struct passwd *
xgetpwnam(char *pwnam)
{
	struct passwd *p = getpwnam(pwnam);
	
	if (p == NULL) {
		printf("Record not found or an error.\n");
		exit(1);
	}
	
	return p;
}

struct group *
xgetgrnam(char *grnam)
{
	struct group *g = getgrnam(grnam);
	
	if (g == NULL) {
		printf("Record not found or an error.\n");
		exit(1);
	}
	
	return g;
}

void no_opt(void)
{
	printf("uid=%u(%s) gid=%u(%s) groups=", 
	       uid, pw->pw_name, gid, gr2->gr_name);

	while(1) {
		gr = getgrent();
		
		if (gr == NULL) break;

		int i = 0;
		while(gr->gr_mem[i] != NULL) {
			if (*pw->pw_name == *gr->gr_mem[i])
				printf("%u(%s),",gr->gr_gid, gr->gr_name);
			i++;
		}
	}
	
	printf("%u(%s)\n", gid, gr2->gr_name); 	
	endgrent();
}

void opt_G(int q)
{
	if (q == 0) {
		printf("%u", gid);
		
		while(1) {
			gr = getgrent();
						
			if (gr == NULL) break;
			
			int i = 0;
			while(gr->gr_mem[i] != NULL) {
				if (*pw->pw_name == *gr->gr_mem[i])
					printf(" %u",gr->gr_gid);
				
				i++;
			}
		}
	} else {
		printf("%s", gr2->gr_name);
		
		while(1) {
			gr = getgrent();
						
			if (gr == NULL) break;
			
			int i = 0;
			while(gr->gr_mem[i] != NULL) {
				if (*pw->pw_name == *gr->gr_mem[i])
					printf(" %s",gr->gr_name);
				
				i++;
			}
		}
	}
	endgrent();
	printf("\n");	
}

void opt_g(int q)
{
	if (q == 0)
		printf("%u\n", gid);
	else
		printf("%s\n", gr2->gr_name);
}

void opt_u(int q)
{
	if (q == 0)
		printf("%u\n", uid);
	else
		printf("%s\n", pw->pw_name);
}

typedef enum {
	OPT_NOPAR = 1,
	OPT_G = 2,
	OPT_Gn = 3,
	OPT_g = 4,
	OPT_gn = 5,
	OPT_u = 6,
	OPT_un = 7,
	ERR_opt = 8,
	ERR_fopt = 9,
	
} TOK_R;

TOK_R get_opt(int argc, char *argv[], char *string)
{
	char opt;
	TOK_R r;
	char s[4]; 
	int k = 0;
	opterr = 0;
	
	while((opt = getopt(argc, argv, string)) != -1) {
		s[k++] = opt;
		if (k == 4) {
			printf("Invalid format options\n");
			exit(1);
		}
	}
	s[k] = '\0';
	
	if (!strcmp(s, "G"))
		r = OPT_G;
	else if (!strcmp(s, "Gn") || !strcmp(s, "nG"))
		r = OPT_Gn;
	else if (!strcmp(s, "g"))
		r = OPT_g;
	else if (!strcmp(s, "gn") || !strcmp(s, "ng"))
		r = OPT_gn;
	else if (!strcmp(s, "u"))
		r = OPT_u;
	else if (!strcmp(s, "un") || !strcmp(s, "un"))
		r = OPT_un;
	else if (*s == 0)
		r = OPT_NOPAR;
	else if (!strcmp(s, "?"))
		r = ERR_opt;
	else
		r = ERR_fopt;
	
	return r;
}

void opt_tok(TOK_R a)
{
	switch(a) {
		case OPT_G:
			opt_G(0);
			break;
		case OPT_NOPAR:
			no_opt();
			break;
		case OPT_Gn:
			opt_G(1);
			break;
		case OPT_g:
			opt_g(0);
			break;
		case OPT_gn:
			opt_g(1);
			break;
		case OPT_u:
			opt_u(0);
			break;
		case OPT_un:
			opt_u(1);
			break;
		case ERR_opt:
			printf("Invalid option\n");
			break;
		case ERR_fopt:
			printf("Invalid format options\n");
			break;
		default:
			break;
	}		
}

int main(int argc, char *argv[])
{
	TOK_R val = get_opt(argc, argv, "gGrnu");

	if (argv[optind] == NULL) {
		uid = getuid();
		pw = xgetpwuid(uid);
		gid = pw->pw_gid;
		gr2 = xgetgrgid(gid);
	} else {
		pw = xgetpwnam(argv[optind]);
		gr2 = xgetgrnam(argv[optind]);
		uid = pw->pw_uid;
		gid = gr2->gr_gid;
	}

	opt_tok(val);
  return EXIT_SUCCESS;
}